// Copyright 2022 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

#include "go_asm.h"
#include "textflag.h"

// Register map
//
// R4: ptr
// R5: n
// R6: ptrend
// R7: tmp
// R8: tmp

// Algorithm:
//
// 1. when count <= 64 bytes, memory alignment check is omitted.
// The handling is divided into distinct cases based on the size
// of count: clr_0, clr_1, clr_2, clr_3, clr_4, clr_5through7,
// clr_8, clr_9through16, clr_17through32, and clr_33through64.
//
// 2. when count > 64 bytes, memory alignment check is performed.
// Unaligned bytes are processed first (that is, 8-(ptr&7)), and
// then a 64-byte loop is executed to zero out memory.
// When the number of remaining bytes not cleared is n < 64 bytes,
// a tail processing is performed, invoking the corresponding case
// based on the size of n.
//
//    ptr           newptr                           ptrend
//     |               |<----count after correction---->|
//     |<-------------count before correction---------->|
//     |<--8-(ptr&7)-->|               |<---64 bytes--->|
//     +------------------------------------------------+
//     |   Head        |      Body     |      Tail      |
//     +---------------+---------------+----------------+
//    newptr = ptr - (ptr & 7) + 8
//    count = count - 8 + (ptr & 7)

// func memclrNoHeapPointers(ptr unsafe.Pointer, n uintptr)
TEXT runtime·memclrNoHeapPointers<ABIInternal>(SB),NOSPLIT,$0-16
	// <=64 bytes, clear directly, not check aligned
generic_small:
	ADDV	R4, R5, R6
	BEQ	R4, R6, clr_0
	MOVV	$2, R7
	BLT	R5, R7, clr_1
	MOVV	$3, R7
	BLT	R5, R7, clr_2
	MOVV	$4, R7
	BLT	R5, R7, clr_3
	MOVV	$5, R7
	BLT	R5, R7, clr_4
	MOVV	$8, R7
	BLT	R5, R7, clr_5_7
	MOVV	$9, R7
	BLT	R5, R7, clr_8
	MOVV	$17, R7
	BLT	R5, R7, clr_9_16
	MOVV	$33, R7
	BLT	R5, R7, clr_17_32
	MOVV	$65, R7
	BLT	R5, R7, clr_33_64

lasx_large:
	MOVBU	internal∕cpu·Loong64+const_offsetLOONG64HasLASX(SB), R7
	BEQ	R7, lsx_large

	// X0 = 0
	XVMOVQ	R0, X0.V4

	// check 32-byte alignment
	AND	$31, R4, R7
	BEQ	R7, lasx_large_aligned
	XVMOVQ	X0, (R4)
	SUBV	R7, R4
	ADDV	R7, R5
	SUBV	$32, R5 // newn = n - (32 - (ptr & 31))
	ADDV	$32, R4 // newptr = ptr + (32 - (ptr & 31))

lasx_large_aligned:
	MOVV	$256, R8
	BLT	R5, R8, lasx_small
lasx_large_body:
	XVMOVQ	X0, 0(R4)
	XVMOVQ	X0, 32(R4)
	XVMOVQ	X0, 64(R4)
	XVMOVQ	X0, 96(R4)
	XVMOVQ	X0, 128(R4)
	XVMOVQ	X0, 160(R4)
	XVMOVQ	X0, 192(R4)
	XVMOVQ	X0, 224(R4)
	SUBV	$256, R5
	ADDV	$256, R4
	BGE	R5, R8, lasx_large_body

lasx_small:
	MOVV	$32, R8
	BLT	R5, R8, generic_small
lasx_small_body:
	XVMOVQ	X0, (R4)
	SUBV	$32, R5
	ADDV	$32, R4
	BGE     R5, R8, lasx_small_body
lasx_tail:
	JMP	generic_small

lsx_large:
	MOVBU	internal∕cpu·Loong64+const_offsetLOONG64HasLSX(SB), R7
	BEQ	R7, generic_large

	// V0 = 0
	VMOVQ	R0, V0.V2

	// check 16-byte alignment
	AND	$15, R4, R7
	BEQ	R7, lsx_large_aligned
	VMOVQ	V0, (R4)
	SUBV	R7, R4
	ADDV	R7, R5
	SUBV	$16, R5 // newn = n - (16 - (ptr & 15))
	ADDV	$16, R4 // newptr = ptr + (16 - (ptr & 15))

lsx_large_aligned:
	MOVV	$128, R8
	BLT	R5, R8, lsx_small
lsx_large_body:
	VMOVQ	V0, 0(R4)
	VMOVQ	V0, 16(R4)
	VMOVQ	V0, 32(R4)
	VMOVQ	V0, 48(R4)
	VMOVQ	V0, 64(R4)
	VMOVQ	V0, 80(R4)
	VMOVQ	V0, 96(R4)
	VMOVQ	V0, 112(R4)
	SUBV	$128, R5
	ADDV	$128, R4
	BGE     R5, R8, lsx_large_body

lsx_small:
	MOVV	$16, R8
	BLT	R5, R8, generic_small
lsx_small_body:
	VMOVQ	V0, (R4)
	SUBV	$16, R5
	ADDV	$16, R4
	BGE     R5, R8, lsx_small_body
lsx_tail:
	JMP	generic_small

generic_large:
	// n > 64 bytes, check aligned
	AND	$7, R4, R7
	BEQ	R7, generic_large_aligned
	MOVV	R0, (R4)
	ADDV	R7, R5
	SUBV	R7, R4
	SUBV	$8, R5	// newn = n - (8 - (ptr & 7))
	ADDV	$8, R4	// newptr = ptr + (8 - (ptr & 7))

generic_large_aligned:
	MOVV    $65, R7
	BLT	R5, R7, generic_small
generic_large_body:
	MOVV	R0, (R4)
	MOVV	R0, 8(R4)
	MOVV	R0, 16(R4)
	MOVV	R0, 24(R4)
	MOVV	R0, 32(R4)
	MOVV	R0, 40(R4)
	MOVV	R0, 48(R4)
	MOVV	R0, 56(R4)
	SUBV	$64, R5
	ADDV	$64, R4
	BGE	R5, R7, generic_large_body
generic_tail:
	JMP	generic_small

clr_33_64:
	MOVV	R0, (R4)
	MOVV	R0, 8(R4)
	MOVV	R0, 16(R4)
	MOVV	R0, 24(R4)
	MOVV	R0, -32(R6)
	MOVV	R0, -24(R6)
	MOVV	R0, -16(R6)
	MOVV	R0, -8(R6)
	RET

clr_17_32:
	MOVV	R0, (R4)
	MOVV	R0, 8(R4)
	MOVV	R0, -16(R6)
	MOVV	R0, -8(R6)
	RET
clr_9_16:
	MOVV	R0, (R4)
	MOVV	R0, -8(R6)
	RET
clr_8:
	MOVV	R0, (R4)
	RET
clr_5_7:
	MOVW	R0, (R4)
	MOVW	R0, -4(R6)
	RET
clr_4:
	MOVW	R0, (R4)
	RET
clr_3:
	MOVH	R0, (R4)
	MOVB	R0, 2(R4)
	RET
clr_2:
	MOVH	R0, (R4)
	RET
clr_1:
	MOVB	R0, (R4)
	RET
clr_0:
	RET
